unit gCodeMain;
  (* syntax checker *)

  interface
    uses
      AdvSpin  (* tAdvSpinEdit *),
      Classes  (* *),
      ComCtrls (* tPageControl, tTabSheet *),
      Controls (* tCreateParams, crHourglass, tCursor, Handle *),
      GCodeDM  (* tCharSet, tLexList, tLexMolecule, tM98PList, tOhList *),
      Forms    (* Application, tForm, Screen *),
      Mask     (* used by AdvSpin *),
      Messages (* EM_LINESCROLL, tMessage, WM_DROPFILES *),
      MyUtils  (* AskYesNo2, UM_EnsureRestored *),
      StdCtrls (* tButton, tCheckBox, tEdit, tGroupBox, tLabel, tListBox,
                    tMemo *),
      Windows  (* tHandle, IsIconic, SetForegroundWindow *);

    type
      tfmGCode = class(tForm)
          pgctlGCode: tPageControl;
            tabListBoxes: tTabSheet;
             spinAssignments: tAdvSpinEdit;
             spinAXYZ:        tAdvSpinEdit;
             spinComments:    tAdvSpinEdit;
             spinFSTetc:      tAdvSpinEdit;
             spinGCodes:      tAdvSpinEdit;
             spinLines:       tAdvSpinEdit;
             spinMCodes:      tAdvSpinEdit;
             spinOhLines:     tAdvSpinEdit;
             btnAnnotateAll:      tButton;
              btnClearAnnotations: tButton;
              btnCopyToClipboard:  tButton;
              btnSave:             tButton;
              ckbxAllowIsolatedG0G1: tCheckBox;
              ckbxShowAllLines:      tCheckBox;
              ckbxDebug:             tCheckBox;
              grpFindFile: tGroupBox;
                btnBrowse:   tButton;
                btnOpenFile: tButton;
                edFileName: tEdit;
              lbAnnotated: tLabel;
              lbNNumbers:  tLabel;
              lbOriginal:  tLabel;
              lbxAnnotated: tListBox;
              memoOriginal: tMemo;
            tabNotes: tTabSheet;
              memoNotes: tMemo;
            tabVariables: tTabSheet;
              btnSaveVarsToFile: tButton;
              lbAssigned:          tLabel;
              lbAssignNoReference: tLabel;
              lbReferenced:        tLabel;
              lbReferenceNoAssign: tLabel;
              lbxAssignedNoReference:   tListBox;
              lbxReferencedNotAssigned: tListBox;
              lbxVarsAssigned:          tListBox;
              lbxVarsUsed:              tListBox;

          procedure btnAnnotateAllClick     (Sender: tObject);
          procedure btnBrowseClick          (Sender: tObject);
          procedure btnClearAnnotationsClick(Sender: tObject);
          procedure btnCopyToClipboardClick (Sender: tObject);
          procedure btnOpenFileClick        (Sender: tObject);
          procedure btnSaveClick            (Sender: tObject);
          procedure btnSaveVarsToFileClick  (Sender: tObject);
          procedure FormDestroy             (Sender: tObject);
          procedure FormShow                (Sender: tObject);
          procedure lbxAnnotatedClick       (Sender: tObject);

          procedure FormCloseQuery(Sender: tObject; var CanClose: boolean);
        private
          ErrorsFound, InOhSub, NonPercentFound, PercentStopper: boolean;
          LineOfInterest: string;
          Parsed: string;
          procedure AnnotateThisLine(const N: integer; const ThisLine: string);
          procedure LogVariable(const VarName: string; const OnLHS: boolean);
          procedure MakeAnnotation(const A: string; LineNumber: integer = 0);
          procedure UMEnsureRestored(var Msg: tMessage); message UM_EnsureRestored;
        protected
          procedure CreateParams(var Params: tCreateParams); override;
          procedure WMDropFiles(var Msg: tMessage); message WM_DROPFILES;
      end (* tfmGCode *);

    var
      fmGCode: tfmGCode;

  implementation
    uses
      Clipbrd     (* Clipboard *),
      Dialogs     (* ofOverwritePrompt, ofShowHelp. tOpenDialog, tSaveDialog *),

      MyConsts    (* AT, BLANK, CIRCUMFLEX, CRLF, EQUALS, GREATER, LBRACKET,
                       LESS, LPAR, MINUS, nullstring, PERCENT, PERIOD, PLUS,
                       RBRACKET, RPAR, SEMICOLON, SHARP, SLASH, STAR, TAB,
                       ULINE *),
      ShellAPI    (* DragAcceptFiles, DragFinish, DragQueryFile *),
      SysUtils    (* EConvertError, ExtractFileExt, Format, LowerCase,
                       rfReplaceAll, StrCopy, StringReplace, StrToFloat,
                       StrToInt, Trim *);
      {$R *.dfm}

    const
      NumberSeparator = '.  ';
      
    var
      OhList: tOhList;
      M98PList: tM98PList;

{ **************** infrastructure *********************************************}
    procedure tfmGCode.CreateParams(var Params: tCreateParams);
      (* see http://www.delphidabbler.com/articles?article=13&part=2 *)
    
      begin (* tfmGCode.CreateParams *)
        inherited CreateParams(Params);
        StrCopy(Params.WinClassName, '20170709 1016');
      end   (* tfmGCode.CreateParams *);

    procedure tfmGCode.FormCloseQuery(Sender: tObject; var CanClose: Boolean);
      (* don't close with dirty memo *)
    
      begin (* tfmGCode.FormCloseQuery *)
        CanClose := true;
        if memoOriginal.Modified then
          CanClose := AskYesNo2('Close without saving file?')
      end   (* tfmGCode.FormCloseQuery *);

    procedure tfmGCode.FormDestroy(Sender: tObject);
      (* close down *)
    
      begin (* tfmGCode.FormDestroy *)
        DragAcceptFiles(Handle, false)
      end   (* tfmGCode.FormDestroy *);

    procedure tfmGCode.FormShow(Sender: tObject);
      (* initialize *)

      procedure EvadeDisplayBug;
        (* The resizing/anchor mechanism seems to be flawed, and
        without a little help here, on some platforms one sees wide, useless
        margins below and to the right of the PageControl. *)

        const
          DesignTimeFormHeight = 1035;
          DesignTimeFormWidth  = 1440;
          DesignTimePgctlHeight = 920;
          DesignTimePgctlWidth = 1411;

        begin (* EvadeDisplayBug *)
          with pgctlGCode do
            begin (* pgctlBoxes3 *)
              Height := self.Height - (DesignTimeFormHeight - DesignTimePgctlHeight);
              Width  := self.Width  - (DesignTimeFormWidth  - DesignTimePgctlWidth)
            end   (* pgctlBoxes3 *)
        end   (* EvadeDisplayBug *);

      begin (* tfmGCode.FormShow *)
        EvadeDisplayBug;
        lbNNumbers.Visible := false;
        pgctlGCode.ActivePage := tabListBoxes;
        DragAcceptFiles(Handle, true)
      end   (* tfmGCode.FormShow *);

    procedure tfmGCode.UMEnsureRestored(var Msg: tMessage);
      (* wake up when user tries to start another instance *)

      begin (* tfmGCode.UMEnsureRestored *)
        if IsIconic(Application.Handle) then
            Application.Restore;
        if not Visible then
          Visible := true;
        Application.BringToFront;
        SetForegroundWindow(self.Handle)
      end   (* tfmGCode.UMEnsureRestored *);

    procedure tfmGCode.WMDropFiles(var Msg: tMessage);
      (* accept a file, dragged from Explorer *)

      const
        ReturnTheCount = $FFFFFFFF;
        NoFileNameBuffer = nil;
        SizeOfNonBuffer = 0;

      var
        hDrop: tHandle;
        FileCount, NameLength: integer;
        S: string;

      begin (* tfmGCode.WMDropFiles *)
        hDrop := Msg.wParam;
        FileCount := DragQueryFile(hDrop, ReturnTheCount, NoFileNameBuffer, SizeOfNonBuffer);

        if FileCount >= 1 then
          begin (* take just the first (zeroth) *)
            NameLength:= DragQueryFile(hDrop, 0, NoFileNameBuffer, SizeOfNonBuffer) + 1;
            SetLength(S, NameLength);
            DragQueryFile(hDrop, 0, pointer(S), NameLength);
            edFileName.Text := S
          end   (* take just the first (zeroth) *);

        DragFinish(hDrop)
      end   (* tfmGCode.WMDropFiles *);

{ **************** action *****************************************************}
    procedure tfmGCode.AnnotateThisLine(const N: integer; const ThisLine: string);
      (* all the work done here *)

      var
        S: string;
        LexList: tLexList;

      var   // debugging
        II: integer;

      procedure LexicalScan;
        (* segment the line into strings *)

        var
           InComment, InNumber, InParameter, MoleculeStarted: boolean;
           ExpressionDepth, ParenDepth: integer;
           TheMolecule: tLexMolecule;

        procedure RealLexicalScan;
          (* separate to shorten the parent *)

          var
            C: char;
            I: integer;

          procedure BadIfInParameterName;
            (* complain if C in parameter name *)

            begin (* BadIfInParameterName *)
              if InParameter then
                MakeAnnotation(Format('"%s" unexpected inside parameter name', [C]))
            end   (* BadIfInParameterName *);

          procedure Singleton;
            (* C needs its own molecule *)

            begin (* Singleton *)
              if MoleculeStarted then
                LexList.Add(TheMolecule);
              TheMolecule := tLexMolecule.CreateForChar(C, I);
              LexList.Add(TheMolecule);
              InNumber := false;
              InParameter := false;
              MoleculeStarted := false
            end   (* Singleton *);

          begin (* RealLexicalScan *)
            for I := 1 to length(S) do
              begin (* process S[I] *)
                C := S[I];
                if InComment then
                  case C of
                    LPAR: inc(ParenDepth);

                    RPAR: if ParenDepth = 1 then
                            InComment := false
                          else
                            dec(ParenDepth);

                     else continue
                    end (* case on C *)
                else
                  case C of
                     BLANK, TAB: continue;

                     '0' .. '9',
                         PERIOD: begin (* numeral or DP *)
                                   if InNumber or InParameter then
                                     TheMolecule.AddChar(C)
                                   else
                                     begin (* new number starting *)
                                       if MoleculeStarted then
                                         LexList.Add(TheMolecule);
                                       TheMolecule := tLexMolecule.CreateForChar(C, I);
                                       MoleculeStarted := true;
                                       InNumber := true
                                     end   (* new number starting *)
                                 end   (* numeral or DP *);

                             AT,
                     CIRCUMFLEX,
                     'a' .. 'z',
                          ULINE: if InNumber then
                                   begin (* number is finished *)
                                     LexList.Add(TheMolecule);
                                     InNumber := false;
                                     TheMolecule := tLexMolecule.CreateForChar(C, I);
                                     MoleculeStarted := true
                                   end   (* number is finished *)
                                 else if MoleculeStarted then
                                   TheMolecule.AddChar(C)
                                 else
                                   begin (* started a string *)
                                     TheMolecule := tLexMolecule.CreateForChar(C, I);
                                     MoleculeStarted := true
                                   end   (* started a string *);

                         EQUALS,
                          MINUS,
                           PLUS,
                          SHARP,
                          SLASH: begin (* punctuation *)
                                   BadIfInParameterName;
                                   Singleton
                                 end   (* punctuation *);

                        GREATER: begin (* should end parameter name *)
                                   if InParameter then
                                     InParameter := false
                                   else
                                     MakeAnnotation(Format('"%s" unexpected at position %d', [C, I]));
                                   Singleton;
                                 end   (* should end parameter name *);

                       LBRACKET: begin (* deeper nesting *)
                                   BadIfInParameterName;
                                   inc(ExpressionDepth);
                                   Singleton
                                 end   (* deeper nesting *);

                           LESS: begin (* start parameter name *)
                                   BadIfInParameterName;
                                   Singleton;
                                   InParameter := true
                                 end   (* start parameter name *);

                           LPAR: begin (* start comment *)
                                   ParenDepth := 1;
                                   InComment := true;
                                   spinComments.Value := spinComments.Value + 1
                                 end   (* start comment *) ;

                       RBRACKET: begin (* shallower nesting *)
                                   BadIfInParameterName;
                                   dec(ExpressionDepth);
                                   Singleton
                                 end   (* shallower nesting *);

                      SEMICOLON: begin (* comment *)
                                   spinComments.Value := spinComments.Value + 1;
                                   break
                                 end   (* comment *);

                           STAR: begin (* ** for exponentiation *)
                                   BadIfInParameterName;
                                   Singleton;
                                   with LexList do
                                     if (Count > 1) and (Items[Count-2].TheString = STAR) then
                                       begin (* change to ** *)
                                         Items[Count-2].TheString := '**';
                                         LexList.Delete(Count-1)
                                       end   (* change to ** *)
                                 end   (* ** for exponentiation *)

                           else  MakeAnnotation(Format('"%s" unexpected at position %d', [C, I]))
                  end (* case *);
              end   (* process S[I] *);

            if MoleculeStarted then
              LexList.Add(TheMolecule);
            if InComment then
              MakeAnnotation('unclosed comment');
            if InParameter then
              MakeAnnotation('parameter name not terminated');
            if ExpressionDepth > 0 then
              MakeAnnotation('unmatched left bracket')
            else if ExpressionDepth < 0 then
              MakeAnnotation('superfluous right bracket')
          end   (* RealLexicalScan *);

        begin (* LexicalScan *)
          MoleculeStarted := false;
          InComment := false;
          InNumber := false;
          InParameter := false;
          ExpressionDepth := 0;
          S := LowerCase(ThisLine);
          ParenDepth := 0;
          if (Trim(ThisLine) = PERCENT) then
            begin (* ignore if at very beginning *)
              if NonPercentFound then
                PercentStopper := true
            end   (* ignore if at very beginning *)
          else
            RealLexicalScan
        end   (* LexicalScan *);

      procedure Parse;
        (* work on LexList *)

        var
          NOK, OhOK: boolean;
          GGroupUsed, MGroupUsed: array[0 .. 15] of boolean;
          I, LexListRunner: integer;

        procedure AnnotateAndBreak(const S: string);
          (* shorthand *)

          begin (* AnnotateAndBreak *)
            MakeAnnotation(S);
            LexListRunner := LexList.Count + 1
          end   (* AnnotateAndBreak *);

        procedure CollectPhrase;
          (* possible phrases are
               F
               GCode
               M
               N (if NOK)
               Oh (if OhOK)
               S
               T
               <parameter> = <expression>
                *)

          var
            AssignmentLHS: boolean;

          function CollectCharacter(const C: char): boolean; forward;
          function CollectExpression: boolean;               forward;
          function CollectNumber: boolean;                   forward;
          function CollectParameter: boolean;                forward;
          function CollectParameterGreater: boolean;         forward;
          function CollectTermsRBracket: boolean;            forward;
          function SuccessAdvancing: boolean;                forward;

          function CollectAssignmentPhrase: boolean;
            (* current item is SHARP.  coming are
                 integer or [expression] or <parameter name>
                 =
                 -number or number or [expression]
            *)

            function CollectBracketedExpression: boolean;
              (* used in LHS & RHS *)

              begin (* CollectBracketedExpression *)
                Result := false;
                if CollectCharacter(LBRACKET) then
                  Result := CollectTermsRBracket
              end   (* CollectBracketedExpression *);

            begin (* CollectAssignmentPhrase *)
              AssignmentLHS := true;
              spinAssignments.Value := spinAssignments.Value + 1;
              Result := false;
              if CollectParameter then
                if CollectCharacter(EQUALS) then
                  begin (* LHS is over *)
                    AssignmentLHS := false;
                    Result := CollectExpression
                  end   (* LHS is over *)
            end   (* CollectAssignmentPhrase *);

          function CollectCharacter(const C: char): boolean;
            (* separate just for clarity *)

            var
              StringFound: string;

            begin (* CollectCharacter *)
              Result := false;
              if SuccessAdvancing then
                begin (* not off the end *)
                  StringFound := LexList.GetLM(LexListRunner-1).TheString;
                  if (length(StringFound) = 1) and (StringFound[1] = C) then
                    Result := true
                  else
                    AnnotateAndBreak(Format('"%s" found where "%s" expected',
                                            [StringFound, C]))
                end   (* not off the end *);
            end   (* CollectCharacter *);

          function CollectExpression: boolean;
            (* expression is
                 -number
                 number
                 #number
                 #<parameter>
                 #[expression]
                 [term]
                 [term {infix term}*]
             *)
            (* non-recursive parser; just demands alternating terms & operators *)

            begin (* CollectExpression *)
              Result := false;
              if SuccessAdvancing then
                with LexList.GetLM(LexListRunner-1) do
                  case TheString[1] of
                      LBRACKET: Result := CollectTermsRBracket;
                         MINUS: Result := CollectNumber;
                    '0' .. '9',
                        PERIOD: Result := true;
                         SHARP: Result := CollectParameter;
                           else AnnotateAndBreak(Format('unexpected "%s" at position %d',
                                            [TheString, StartIndex]))
                    end (* case on StringFound[1] *)
            end   (* CollectExpression *);

          procedure CollectGPhrase;
            (* current string is G.  Collect Gnumber and its arguments *)

            var
              RealGNumber: double;

            procedure PlausibleGCode;
              (* extra procedure to shorten CollectGPhrase *)

              var
                Found, G0001, G0203, Polar, SArgOK: boolean;
                ArgsFound, HaveAtLeastOne, MayHaveArgs, NeededArgs,
                  NeedAtLeastOne: tCharSet;
                I, ThisGroup: integer;
                MissingArgs, UnexpectedArgs: string;

              function CollectArg: boolean;
                (* try to collect the next two items *)

                const
                  UsualArgs = [AT, CIRCUMFLEX, 'a' .. 'd', 'i' .. 'l',
                               'p' .. 'r', 'u' .. 'z'];

                var
                  PossibleLetters: tCharSet;
                  Letter: string;

                begin (* CollectArg *)
                  Result := false;
                  PossibleLetters := UsualArgs;
                  if SArgOK then
                    PossibleLetters := PossibleLetters + ['s'];

                  inc(LexListRunner);
                  if LexListRunner <= LexList.Count then
                    with LexList.GetLM(LexListRunner-1) do
                      begin (* single letter argument code *)
                        Letter := LowerCase(TheString);
                        if length(Letter) > 1 then
                          MakeAnnotation(Format('unexpected %s after %s at position %d',
                                                [copy(TheString, 2, maxint),
                                                 copy(TheString, 1, 1),
                                                 StartIndex]))
                        else if (Letter[1] in PossibleLetters) and CollectExpression then
                          begin (* found one argument *)
                            ArgsFound := ArgsFound + [Letter[1]];
                            Result := true
                          end   (* found one argument *)
                        else
                          dec(LexListRunner) // unpeek
                      end   (* single letter argument code *)
                end   (* CollectArg *);

              begin (* PlausibleGCode *)
                ArgsFound := [];
                Found := false;
                for I := 1 to NGCodes do
                  with GCode[I] do
                    if abs(Number - RealGNumber) < 0.001 then
                      begin (* found it *)
                        Found := true;
                        ThisGroup := Group;
                        NeededArgs := MustHave;
                        NeedAtLeastOne := AtLeastOne;
                        MayHaveArgs := MayHave;
                        SArgOK := (Number = 96);
                        G0001 := ((Number = 0) or (Number = 1));
                        G0203 := ((Number = 2) or (Number = 3));

                        if GGroupUsed[ThisGroup] then
                          AnnotateAndBreak(Format('G%.1g incompatible with gCode used earlier', [RealGNumber]))
                        else
                          begin (* nondegenerate *)
                            if ThisGroup <> 0 then
                              GGroupUsed[ThisGroup] := true;
                            while CollectArg do;

                            MissingArgs := dmGCode.CharSetToString(NeededArgs - ArgsFound);
                            HaveAtLeastOne := NeedAtLeastOne * ArgsFound;
                            UnexpectedArgs := dmGCode.CharSetToString(ArgsFound - (NeededArgs+NeedAtLeastOne+MayHaveArgs));
                            Polar := ((AT in ArgsFound) or (CIRCUMFLEX in ArgsFound));

                            if Polar and (('x' in ArgsFound) or ('y' in ArgsFound)) then
                              AnnotateAndBreak('mixed Cartesian and polar arguments')
                            else if MissingArgs <> nullstring then
                              AnnotateAndBreak(Format('G%.1g needs %s argument(s)',
                                                      [RealGNumber, MissingArgs]))
                            else if (NeedAtLeastOne <> []) and (HaveAtLeastOne = []) then
                              if G0001 and ckbxAllowIsolatedG0G1.Checked then
                                (* ignore it *)
                              else
                                AnnotateAndBreak(Format('G%.1g needs at least one of %s as argument(s)',
                                                        [RealGNumber, dmGCode.CharSetToString(NeedAtLeastOne)]))
                            else if UnexpectedArgs <> nullstring then
                              AnnotateAndBreak(Format('unexpected %s argument(s) given to G%.1g',
                                                      [UnexpectedArgs, RealGNumber]))
                            else if G0203 then
                              if ('r' in ArgsFound) then
                                begin (* radius format *)
                                  if (['x', 'y'] * ArgsFound) = [] then
                                    AnnotateAndBreak(Format('radius-format G%2d needs at least one of [x, y] as argument(s)',
                                                        [trunc(RealGNumber)]))
                                end   (* radius format *)
                              else if ['i', 'j'] * ArgsFound = [] then
                                AnnotateAndBreak(Format('center-format G%2d needs at least one of [i, j] as argument(s)',
                                                        [trunc(RealGNumber)]))
                          end   (* nondegenerate *);
                        break
                      end   (* found it *);

                if not Found then
                  AnnotateAndBreak(Format('G%.1g not recognized', [RealGNumber]))
              end   (* PlausibleGCode *);

            begin (* CollectGPhrase *)
              spinGCodes.Value := spinGCodes.Value + 1;
              if SuccessAdvancing then
                with LexList.GetLM(LexListRunner-1) do
                  if TheString[1] in ['0' .. '9', PERIOD] then
                    begin (* putative GNumber *)
                      try    (* float conversion is fragile *)
                        RealGNumber := StrToFloat(TheString)
                      except (* float conversion is fragile *)
                        on E: EConvertError do
                          RealGNumber := -1
                      end    (* float conversion is fragile *);

                      if (RealGNumber < 0) or (RealGNumber > 98) then
                        AnnotateAndBreak('G-number bad')
                      else
                        PlausibleGCode
                    end (* putative Gnumber *)
            end   (* CollectGPhrase *);

          function CollectInfixOp: boolean;
            (* if not found, unpeek *)

            type
              tInfixOpArray = array[1 .. 14] of string;

            const
              InfixOps: tInfixOpArray = ('and', 'eq',  'ge',  'gt', 'le',
                                         'lt',  MINUS, 'mod', 'ne', 'or',
                                         PLUS,  SLASH, STAR,  'xor');

            var
              Found: boolean;
              I: integer;

            begin (* CollectInfixOp *)
              Found := false;
              if LexListRunner < LexList.Count then
                begin (* possible to advance *)
                  inc(LexListRunner);
                  for I := low(InfixOps) to high(InfixOps) do
                    if LowerCase(LexList.GetLM(LexListRunner-1).TheString) = InfixOps[I] then
                      begin (* found *)
                        Found := true;
                        break
                      end   (* found *)
                end   (* possible to advance *);

              if not Found then
                dec(LexListRunner);
              Result := Found
            end   (* CollectInfixOp *);

          function CollectInteger: integer;
            (* next item should be integer *)

            begin (* CollectInteger *)
              Result := -1;
              if SuccessAdvancing then
                with LexList.GetLM(LexListRunner-1) do
                  try    (* conversion is fragile *)
                    Result := StrToInt(TheString)
                  except (* conversion is fragile *)
                    AnnotateAndBreak(Format('%s found where integer expected at position %d',
                                            [TheString, StartIndex]));
                  end    (* conversion is fragile *)
            end   (* CollectInteger *);

          function CollectNumber: boolean;
            (* next item should be number *)

            begin (* CollectNumber *)
              Result := false;
              if SuccessAdvancing then
                with LexList.GetLM(LexListRunner-1) do
                  case TheString[1] of
                    '0' .. '9',
                        PERIOD: Result := true
                           else AnnotateAndBreak(Format('%s found where number expected at position %d',
                                                        [TheString, StartIndex]))
                    end (* case *)
            end   (* CollectNumber *);

          procedure CollectMPhrase;
            (* Current string is M.  Next should be integer *)

            var
              MNumber: integer;

            procedure PlausibleMCode;
              (* may be done, but check group, and may need arguments *)
            
              var
                Found: boolean;
                ArgsFound, MayHaveArgs, NeededArgs: tCharSet;
                I, ArgNumberIndex, ThisGroup: integer;
                MissingArgs, M98PString, UnexpectedArgs: string;

              function CollectArg: boolean;
                (* try to collect the next two items *)

                const
                  PossibleLetters = ['l', 'p', 'q'];

                var
                  Letter: string;

                begin (* CollectArg *)
                  Result := false;

                  inc(LexListRunner);
                  if LexListRunner <= LexList.Count then
                    with LexList.GetLM(LexListRunner-1) do
                      begin (* single letter argument code *)
                        Letter := LowerCase(TheString);
                        if length(Letter) > 1 then
                          MakeAnnotation(Format('unexpected %s after %s at position %d',
                                                [copy(TheString, 2, maxint),
                                                 copy(TheString, 1, 1),
                                                 StartIndex]))
                        else if (Letter[1] in PossibleLetters) and CollectExpression then
                          begin (* found one argument *)
                            ArgsFound := ArgsFound + [Letter[1]];
                            ArgNumberIndex := LexListRunner; // assumes P-number is simple number
                            Result := true
                          end   (* found one argument *)
                        else
                          dec(LexListRunner)
                      end   (* single letter argument code *)
                end   (* CollectArg *);

              begin (* PlausibleMCode *)
                ArgsFound := [];
                Found := false;
                for I := 1 to NMCodes do
                  with MCode[I] do
                    if Number = MNumber then
                      begin (* found it *)
                        Found := true;
                        ThisGroup := Group;
                        NeededArgs := MustHave;
                        MayHaveArgs := MayHave;

                        if MGroupUsed[ThisGroup] then
                          AnnotateAndBreak(Format('M%2d incompatible with MCode used earlier', [MNumber]))
                        else
                          begin (* nondegenerate *)
                            if ThisGroup <> 0 then
                              MGroupUsed[ThisGroup] := true;
                            while CollectArg do;

                            MissingArgs := dmGCode.CharSetToString(NeededArgs - ArgsFound);
                            UnexpectedArgs := dmGCode.CharSetToString(ArgsFound - (NeededArgs+MayHaveArgs));

                            if MissingArgs <> nullstring then
                              AnnotateAndBreak(Format('M%2d needs %s argument(s)',
                                                      [MNumber, MissingArgs]))
                            else if UnexpectedArgs <> nullstring then
                              AnnotateAndBreak(Format('unexpected %s argument(s) given to M%2d',
                                                      [UnexpectedArgs, MNumber]));
                            if MNumber = 98 then
                              begin (* try for M98P item *)
                                M98PString := LexList.GetLM(ArgNumberIndex-1).TheString;
                                if M98PString[1] in ['0' .. '9'] then
                                  M98PList.Add(tM98PItem.Create(N, StrToInt(M98PString)))
                              end   (* try for M98P item *)
                          end   (* nondegenerate *);
                        break
                      end   (* found it *);

                if not Found then
                  AnnotateAndBreak(Format('M%2d not recognized', [MNumber]))
              end   (* PlausibleMCode *);

            begin (* CollectMPhrase *)
              spinMCodes.Value := spinMCodes.Value + 1;
              if LexListRunner = LexList.Count then
                AnnotateAndBreak('line ended prematurely');
              inc(LexListRunner);
              with LexList.GetLM(LexListRunner-1) do
                if TheString[1] in ['0' .. '9'] then
                  begin (* putative M-number *)
                    try    (* conversion is fragile *)
                      MNumber := StrToInt(TheString)
                    except (* conversion is fragile *)
                      on E: EConvertError do
                        MNumber := -1
                    end    (* conversion is fragile *);

                    if (MNumber < 0) or (MNumber > 99) then
                      AnnotateAndBreak('M-number not recognized')
                    else
                      PlausibleMCode
                  end   (* putative M-number *)
            end   (* CollectMPhrase *);

          procedure CollectOh;
            (* maintain OhList.  Possible Oh entries are
                 break
                 call
                 continue
                 do
                 else
                 elseif
                 endif
                 endrepeat
                 endsub
                 endwhile
                 if
                 null
                 repeat
                 return
                 sub
                 while
                  *)

            var
              FileNameGiven: boolean;
              OhNumber: integer;
              CallEndSub, FileName, NumberOrFileName: string;

            procedure CollectArgumentList;
              (* Collect zerro or more bracketed expressions *)

              var
                Item: string;

              begin (* CollectArgumentList *)
                while LexListRunner < LexList.Count do
                  begin (* try for another *)
                    inc(LexListRunner);
                    Item := LexList.GetLM(LexListRunner-1).TheString;
                    if Item <> LBRACKET then
                      begin (* something, but not another bracketed arg *)
                        AnnotateAndBreak(Format('%s found where bracketed call argument expected',
                                                [Item]));
                        break
                      end   (* something, but not another bracketed arg *)
                    else if not CollectTermsRBracket then
                      break
                  end   (* try for another *)
              end   (* CollectArgumentList *);

            procedure OhNotFileCall;
              (* separate just to shorten parent *)

              var
                GivenVerb, VerbRunner: tOhVerb;
                StringVerb: string;

              begin (* OhNotFileCall *)
                if LexListRunner = LexList.Count then
                  OhList.Add(tOhItem.Create(N, OhNumber, ohNull))
                else
                  begin (*  look for function word *)
                    inc(LexListRunner);
                    GivenVerb := ohNotFound;
                    StringVerb := LexList.GetLM(LexListRunner-1).TheString;
                    for VerbRunner := ohBreak to ohWhile do
                      with OhData[VerbRunner] do
                        if OhString = StringVerb then
                          begin (* found it *)
                            GivenVerb := VerbRunner;
                            case GivenVerb of
                                ohCall: CollectArgumentList;

                              ohElseif,
                                  ohIf,
                              ohRepeat,
                               ohWhile: CollectExpression;

                              ohEndSub: begin (* pop sub depth *)
                                          if InOhSub then
                                            InOhSub := false; // error if not, will be caught later
                                          if LexListRunner < LexList.Count then
                                            CollectExpression
                                        end   (* pop sub depth *);

                              ohReturn: if LexListRunner < LexList.Count then
                                          CollectExpression;

                                 ohSub: if InOhSub then
                                          AnnotateAndBreak('"o <nnnn> sub" may not be nested')
                                        else
                                          InOhSub := true
                              end (* case *);
                            OhList.Add(tOhItem.Create(N, OhNumber, GivenVerb));

                            break
                          end   (* found it *);
                    if GivenVerb = ohNotFound then
                      AnnotateAndBreak(Format('unexpected "%s" in line %d', [StringVerb, N]))
                  end   (*  look for function word *)
              end   (* OhNotFileCall *);

            function PossibleFileName(const S: string): boolean;
              (* check for permissible letters *)

              var
                I: integer;

              begin (* PossibleFileName *)
                Result := true;
                for I := 1 to length(S) do
                  if not (S[I] in [MINUS, ULINE, '0' .. '9', 'a' .. 'z']) then
                    begin (* failure *)
                      Result := false;
                      break
                    end   (* failure *)
              end   (* PossibleFileName *);

            begin (* CollectOh *)
              spinOhLines.Value := spinOhLines.Value + 1;
              FileNameGiven := false;
              OhNumber := 0;
              FileName := nullstring;
              if SuccessAdvancing then
                begin (* might be filecall *)
                  NumberOrFileName := LexList.GetLM(LexListRunner-1).TheString;
                  case NumberOrFileName[1] of
                    '0' .. '9': begin (* back up to collect integer *)
                                  dec(LexListRunner);
                                  OhNumber := CollectInteger;
                                  if OhNumber < 1 then
                                    AnnotateAndBreak('integer expected but not found as o-number')
                                end   (* back up to collect integer *);

                          LESS: begin (* look for string> *)
                                  if SuccessAdvancing then
                                    begin (* here's the file name *)
                                      FileName := LexList.GetLM(LexListRunner-1).TheString;
                                      if PossibleFileName(FileName) then
                                        if CollectCharacter(GREATER) then
                                          FileNameGiven := true
                                    end   (* here's the file name *);
                                  if not FileNameGiven then
                                    AnnotateAndBreak('file name expected but not found')
                                end   (* look for string> *)

                           else AnnotateAndBreak('Oh line has neither number nor file name')
                    end (* case on item just after Oh *)
                end   (* might be filecall *);

              if LexListRunner > LexList.Count then
                (* annotated already *)

              else if not FileNameGiven then
                OhNotFileCall

              else if SuccessAdvancing then
                begin (* o <filename> received, call args to come *)
                  CallEndSub := LexList.GetLM(LexListRunner-1).TheString;
                  if CallEndSub = 'call' then
                    begin (* all set *)
                      CollectArgumentList;
                      OhList.Add(tOhItem.Create(N, OhNumber, ohFileCall, FileName))
                    end   (* all set *)
                  else if CallEndSub = 'endsub' then
                      OhList.Add(tOhItem.Create(N, OhNumber, ohEndSub, FileName))
                  else if CallEndSub = 'sub' then
                      OhList.Add(tOhItem.Create(N, OhNumber, ohSub, FileName))
                  else
                    AnnotateAndBreak(Format('call|end|sub" expected but "%s" found', [CallEndSub]))
                end   (* o <filename> received, call args to come *)
            end   (* CollectOh *);

          function CollectParameter: boolean;
            (* Current item is SHARP.  Next should be
                  integer
                  [expression] or
                  <parametername>
            *)

            var
              StringFound: string;

            begin (* CollectParameter *)
              Result := false;
              if SuccessAdvancing then
                begin (* not off the end *)
                  StringFound := LexList.GetLM(LexListRunner-1).TheString;
                  case StringFound[1] of
                    '0' .. '9': if pos(PERIOD, StringFound) = 0 then
                                  begin (* numerical paramer found *)
                                    LogVariable(SHARP + StringFound, AssignmentLHS);
                                    Result := true
                                  end   (* numerical paramer found *)

                                else
                                  AnnotateAndBreak('parameter number must be integer');
                      LBRACKET: Result := CollectTermsRBracket;
                          LESS: Result := CollectParameterGreater
                           else AnnotateAndBreak('unrecognized parameter')
                      end (* case *)
                end   (* not off the end *)
            end   (* CollectParameter *);

          function CollectParameterGreater: boolean;
            (* string, then ">" *)

            var
              StringFound: string;

            begin (* CollectParameterGreater *)
              Result := false;
              if SuccessAdvancing then
                begin (* not off the end *)
                  StringFound := LowerCase(LexList.GetLM(LexListRunner-1).TheString);
                  case StringFound[1] of
                    'a' .. 'z',
                    '0' .. '9',
                         ULINE: begin (* '#<' processed earlier *)
                                  LogVariable(Format('#<%s>', [StringFound]), AssignmentLHS);
                                  Result := CollectCharacter(GREATER)
                                end   (* '#<' processed earlier *);

                           else AnnotateAndBreak(Format('bad parameter name at position %d',
                                                        [LexList.GetLM(LexListRunner-1).StartIndex]))
                      end (* case *)
                end   (* not off the end *);
            end   (* CollectParameterGreater *);

          function CollectTerm: boolean;
            (* term is
                 -term
                 number
                 function[expression]
                 #number
                 #[expression]
                 #<parameter name>
                 [expression]
             *)

            type
              tFuncArray = array[1 .. 14] of string;

            const
              Functions: tFuncArray = ('atan',   'abs', 'acos', 'asin', 'cos',
                                       'exists', 'exp', 'fix',  'fup',  'ln',
                                       'round',  'sin', 'sqrt', 'tan');
            var
              IsFunction: boolean;
              I: integer;

            begin (* CollectTerm *)
              Result := false;
              if SuccessAdvancing then
                with LexList.GetLM(LexListRunner-1) do
                  case TheString[1] of
                      LBRACKET: Result := CollectTermsRBracket;
                         MINUS: Result := CollectTerm;
                    '0' .. '9',
                        PERIOD: Result := true;
                         SHARP: Result := CollectParameter;
                           else begin (* might be function *)
                                  IsFunction := false;
                                  for I := low(Functions) to High(Functions) do
                                    if TheString = Functions[I] then
                                      begin (* found *)
                                        IsFunction := true;
                                        if CollectCharacter(LBRACKET) then
                                          Result := CollectTermsRBracket;
                                        break
                                      end   (* found *);
                                  if not IsFunction then
                                    AnnotateAndBreak(Format('unexpected "%s" at position %d',
                                                            [TheString, StartIndex]))
                                end   (* might be function *);
                    end (* case *)
            end   (* CollectTerm *);

          function CollectTermsRBracket: boolean;
            (* collect a term and 0 or more of
                 infixop  term
            *)

            var
              KeepTrying: boolean;

            begin (* CollectTermsRBracket *)
              Result := false;
              if CollectTerm then
                begin (* look for 0 or more (infixop & term) *)
                  Result := true;
                  KeepTrying := true;
                  while KeepTrying do
                    if CollectInfixOp then
                      Result := CollectTerm
                    else
                      KeepTrying := false
                end   (* look for 0 or more (infixop & term) *);
              if Result then
                Result := CollectCharacter(RBRACKET)
            end   (* CollectTermsRBracket *);

          function SuccessAdvancing: boolean;
            (* used everywhere *)

            begin (* SuccessAdvancing *)
              Result := false;
              if LexListRunner > LexList.Count then
                AnnotateAndBreak('line ends prematurely')
              else
                begin (* not off the end *)
                  Result := true;
                  inc(LexListRunner)
                end   (* not off the end *)
            end   (* SuccessAdvancing *);

          begin (* CollectPhrase *)
            AssignmentLHS := false;
            with LexList.GetLM(LexListRunner-1) do
              if length(TheString) <> 1 then
                AnnotateAndBreak(Format('unexpected "%s" in "%s"',
                                        [copy(TheString, 2, maxint), TheString]))
              else
                begin (* only N may precede Oh *)
                  case LowerCase(TheString)[1] of
                    'a', 'x',
                    'y', 'z': begin (* axes *)
                                spinAXYZ.Value := spinAXYZ.Value + 1;
                                CollectExpression;
                              end   (* axes *);

                    'f', 'h',
                    's', 't': begin (* FHST *)
                                spinFSTetc.Value := spinFSTetc.Value + 1;
                                CollectExpression;
                              end   (* FHST *);

                         'g': CollectGPhrase;

                         'm': CollectMPhrase;

                         'n': if NOK then
                                begin (* N numbers mainly ignored *)
                                  CollectInteger;
                                  lbNNumbers.Visible := true
                                end   (* N numbers mainly ignored *);

                         'o': if OhOK then
                                CollectOh;

                       SHARP: CollectAssignmentPhrase

                         else AnnotateAndBreak(Format('unexpected "%s" at position %d',
                                                      [TheString, StartIndex]))

                    end (* case on String[1] *);
                  OhOK := (TheString[1] = 'n');
                  NonPercentFound := true
                end   (* only N may precede Oh *)
          end   (* CollectPhrase *);

        begin (* Parse *)
          NOK := true;
          OhOK := true;
          LexListRunner := 1;
          for I := 0 to 15 do
            begin (* nothing seen yet *)
              GGroupUsed[I] := false;
              MGroupUsed[I] := false
            end   (* nothing seen yet *);

          II := 1;
          Parsed := nullstring;
          while II <= LexList.Count do
            begin (* test *)
              Parsed := Parsed + ', ' + LexList.Items[II-1].TheString;
              inc(II)
            end   (* test *);
          Parsed := copy(Parsed, 3, maxint);

          while LexListRunner <= LexList.Count do
            begin (* collect one freestanding gCode phrase *)
              CollectPhrase;
              NOK := false;
              inc(LexListRunner)
            end   (* collect one freestanding gCode phrase *)
        end   (* Parse *);

      begin (* tfmGCode.AnnotateThisLine *)
        if N > 385 then
          ErrorsFound := false;
        LexList := tLexList.Create;
        LineOfInterest := Format('%3d. %s', [N, ThisLine]);
        if ckbxShowAllLines.Checked then
          begin (* need not put it out again *)
            with lbxAnnotated do
              ItemIndex := Items.Add(LineOfInterest);
            LineOfInterest := nullstring;
          end   (* need not put it out again *);

        ErrorsFound := false;
        LexicalScan;
        Parse;
        LexList.Free
      end   (* tfmGCode.AnnotateThisLine *);

      procedure tfmGCode.btnAnnotateAllClick(Sender: tObject);
        (* all the work is done here *)

        var
          I: integer;

        procedure CheckM98PList;
          (* called subroutine must exist *)

          var
            Found: boolean;
            I, J, MPLineNumber, Target: integer;

          begin (* CheckM98PList *)
            with M98PList do
              begin (* M98PList *)
                for I := 1 to Count do
                  begin (* find one target *)
                    MPLineNumber := GetM98P(I-1).LineNumber;
                    Target       := GetM98P(I-1).PNumber;
                    Found := false;
                    if ckbxDebug.Checked then
                      MakeAnnotation(Format('Checking M98 call on line %d',
                                            [MPLineNumber]));
                    with OhList do
                      begin (* OhList *)
                        for J := 1 to Count do
                          if GetOh(J-1).OhNumber = Target then
                            begin (* got it *)
                              Found := true;
                              break
                            end   (* got it *);

                          if not Found then
                            MakeAnnotation(Format('subroutine #%d called from line %d not found',
                                                  [Target, MPLineNumber]))
                      end   (* OhList *)
                  end   (* find one target *)
              end   (* M98PList *)
          end   (* CheckM98PList *);

        procedure CheckOhList;
          (* make sure ifs are matched with endifs, etc. *)

          const
            StartClosed = true;
            AllowNotFound = false;

          var
            BaseIndex, BaseLineNumber, BaseOh, I: integer;
            BaseVerb: tOhVerb;

          function FindOhBase(MustBeFound: boolean = true): integer; forward;
          procedure OhMustBeUnique;                                  forward;

          procedure CloseOpenOh;
            (* close the oH line that opened this OhNumber *)

            begin (* CloseOpenOh *)
              with OhData[BaseVerb] do
                BaseIndex := FindOhBase;
              if BaseIndex >= 0 then
                OhList.GetOh(BaseIndex-1).Closed := true
            end   (* CloseOpenOh *);

        function FindOhBase(MustBeFound: boolean = true): integer;
          (* this is elseif, endif, endwhile, etc *)

          var
            J: integer;

          begin (* FindOhBase *)
            Result := -1;

            J := I-1;
            while J >= 1 do
              begin (* look for open base *)
                with OhList.GetOh(J-1) do
                  if Closed and (BaseVerb <> OhCall) then
                    dec(J)
                  else if (not (TheVerb in OhData[BaseVerb].PossibleBase) or (BaseOh <> OhNumber)) then
                    if BaseVerb in [ohBreak, ohCall, ohContinue, ohWhile] then
                      dec(J)
                    else
                      begin (* mismatch *)
                        MakeAnnotation(Format('"%s" on line %d in range of mismatch on line %d',
                                              [OhData[BaseVerb].OhString, BaseLineNumber, LineNumber]),
                                       LineNumber);
                        break
                      end   (* mismatch *)
                  else
                    begin (* got it *)
                      Result := J;
                      break
                    end   (* got it *)
              end   (* look for open base *);

            if Result < 0 then
              if MustBeFound then
                MakeAnnotation(Format('"%s" on line %d is an orphan',
                                      [OhData[BaseVerb].OhString, BaseLineNumber]),
                               BaseLineNumber)
          end   (* FindOhBase *);

          procedure OhMustBeUnique;
            (* BaseOh must be new *)

            var
              J: integer;

            begin (* OhMustBeUnique *)
              for J := 1 to I-1 do
                with OhList.GetOh(J-1) do
                  if OhNumber = BaseOh then
                    begin (* got it *)
                      MakeAnnotation(Format('O-number %d from line %d reused on line %d',
                                            [BaseOh, LineNumber, BaseLineNumber]),
                                     BaseLineNumber);
                      break
                    end   (* got it *)
            end   (* OhMustBeUnique *);

          begin (* CheckOhList *)
            with OhList do
              begin (* OhList *)
                for I := 1 to Count do
                  with GetOh(I-1) do
                    begin (* see if this Oh closes another *)
                      BaseVerb := TheVerb;
                      BaseOh := OhNumber;
                      BaseLineNumber := LineNumber;
                      if ckbxDebug.Checked then
                        with lbxAnnotated do
                          ItemIndex := Items.Add(Format('Checking o %d %s on line %d',
                                                        [OhNumber, OhData[BaseVerb].OhString, BaseLineNumber]));
                      case BaseVerb of
                           ohBreak,
                        ohContinue,
                            ohCall,
                          ohElseIf: FindOhBase;

                              ohDo,
                              ohIf,
                          ohRepeat,
                             ohSub: OhMustBeUnique;

                           ohEndIf,
                       ohEndRepeat,
                          ohEndSub,
                        ohEndWhile: CloseOpenOh;

                        ohFileCall: MakeAnnotation(Format('call on line %d presumes existence of file ~/subroutines/%s',
                                                          [BaseLineNumber, TheFileName]),
                                                   BaseLineNumber);

                            ohNull: OhMustBeUnique;

                           ohWhile: begin (* could close do *)
                                      BaseIndex := FindOhBase(AllowNotFound);
                                      if BaseIndex >= 0 then
                                        begin (* close do and self *)
                                          GetOh(BaseIndex-1).Closed := true;
                                          Closed := true
                                        end   (* close do and self *)
                                      else
                                        OhMustBeUnique
                                    end   (* could close do *)
                      end (* case *)
                    end   (* see if this Oh closes another *);

                for I := 1 to Count do
                  with GetOh(I-1) do
                    begin (* look for open Ohs *)
                      if not Closed then
                        with lbxAnnotated do
                          ItemIndex := Items.Add(Format('o %d left open', [OhNumber]))
                    end   (* look for open Ohs *)
              end   (* OhList *)
          end   (* CheckOhList *);

        begin (* tfmGCode.btnAnnotateAllClick *)
          InOhSub := false;
          lbNNumbers.Visible := false;
          OhList := tOhList.Create;
          M98PList := tM98PList.Create;
          NonPercentFound := false;
          PercentStopper := false;
          with lbxAnnotated do
            ItemIndex := Items.Add('(checking line by line)');
          spinAssignments.Value := 0;
          spinAXYZ.Value        := 0;
          spinComments.Value    := 0;
          spinFSTetc.Value      := 0;
          spinGCodes.Value      := 0;
          spinLines.Value       := memoOriginal.Lines.Count;
          spinMCodes.Value      := 0;
          spinOhLines.Value     := 0;
          lbxAssignedNoReference.Clear;
          lbxReferencedNotAssigned.Clear;
          lbxVarsAssigned.Clear;
          lbxVarsUsed.Clear;
          with memoOriginal do
            for I := 1 to Lines.Count do
              begin (* one line *)
                AnnotateThisLine(I, copy(Lines[I-1], 6, maxint));
                if PercentStopper then
                  begin (* no more scanning of this file *)
                    with lbxAnnotated do
                      ItemIndex := Items.Add(Format('(EOF marker (%%) found on line %d)',
                                                    [I]));
                    break
                  end   (* no more scanning of this file *)
              end   (* one line *);

          with lbxAnnotated do
            ItemIndex := Items.Add('(checking o <nnnn> structure)');
          CheckOhList;
          with lbxAnnotated do
            ItemIndex := Items.Add('(checking M98 P<nnnn> structure)');
          CheckM98PList;
          with lbxAnnotated do
            ItemIndex := Items.Add('(done)');
          M98PList.Free;
          OhList.Free
        end   (* tfmGCode.btnAnnotateAllClick *);

    procedure tfmGCode.btnBrowseClick(Sender: tObject);
      (* fill the edit box *)

      begin (* tfmGCode.btnBrowseClick *)
        with tOpenDialog.Create(nil) do
          try     (* must free dialog *)
            DefaultExt := 'ncc';
            Filter := 'gCode (*.ncc, *.ngc)|*.ncc;*.nc;*.ngc|all files|*.*';
            Title := 'find gCode file to check';
            if Execute then
              begin (* got it *)
                edFileName.Text := FileName;
                memoOriginal.Clear
              end   (* got it *)
          finally (* must free dialog *)
            Free
          end     (* must free dialog *);
      end   (* tfmGCode.btnBrowseClick *);

    procedure tfmGCode.btnClearAnnotationsClick(Sender: tObject);
      (* clear the annotations box *)
    
      begin (* tfmGCode.btnClearAnnotationsClick *)
        lbxAnnotated.Clear
      end   (* tfmGCode.btnClearAnnotationsClick *);

    procedure tfmGCode.btnCopyToClipboardClick(Sender: tObject);
      (* selected lines or all *)
    
      begin (* tfmGCode.btnCopyToClipboardClick *)
        Clipboard.SetTextBuf(pchar(lbxAnnotated.Items.Text))
      end   (* tfmGCode.btnCopyToClipboardClick *);

    procedure tfmGCode.btnOpenFileClick(Sender: tObject);
      (* load numbered lines into Original listbox *)

      var
        OldCursor: tCursor;
        N: integer;
        Line: string;
        TheText: TextFile;

      begin (* tfmGCode.btnOpenFileClick *)
        OldCursor := Screen.Cursor;
        Screen.Cursor := crHourglass;
        try     (* restore cursor *)
          memoOriginal.Clear;
          lbxAnnotated.Clear;
          N := 0;

          AssignFile(TheText, edFileName.Text);
          Reset(TheText);
          while not System.EOF(TheText) do
            begin (* process one line *)
              readln(TheText, Line);
              inc(N);
              with memoOriginal do
                Lines.Text :=
                  Lines.Text                                                +
                  Format('%3d%s%s',
                         [N,
                          NumberSeparator,
                          StringReplace(Line, TAB, BLANK, [rfReplaceAll])]) +
                  CRLF
            end   (* process one line *);
          CloseFile(TheText);
          memoOriginal.Modified := false
        finally (* restore cursor *)
          Screen.Cursor := OldCursor
        end     (* restore cursor *)
      end   (* tfmGCode.btnOpenFileClick *);

    procedure tfmGCode.btnSaveClick(Sender: tObject);
      (* save the file (without line numbers) *)

      var
        DotPos, I: integer;
        Line: string;
        TheFile: textfile;

      begin (* tfmGCode.btnSaveClick *)
        with tSaveDialog.Create(nil) do
          try     (* must free dialog *)
            DefaultExt := 'ncc';
            FileName := edFileName.Text;
            Options := [ofOverwritePrompt, ofShowHelp];
            Title := 'save file (without line numbers)';
            if Execute then
              begin (* write it out *)
                AssignFile(TheFile, Filename);
                rewrite(TheFile);

                with memoOriginal do
                  for I := 1 to Lines.Count do
                    begin (* one line *)
                      Line := Lines[I-1];
                      DotPos := pos(NumberSeparator, Line);
                      if DotPos > 0 then
                        if DotPos < 6 then
                          Line := copy(Line, DotPos + length(NumberSeparator), maxint);
                      writeln(TheFile, Line)
                    end   (* one line *);

                CloseFile(TheFile);
                memoOriginal.Modified := false
              end   (* write it out *)
          finally (* must free dialog *)
            Free
          end     (* must free dialog *);
      end   (* tfmGCode.btnSaveClick *);

    procedure tfmGCode.lbxAnnotatedClick(Sender: tObject);
      (* try to find the matching line in the original *)

      var
        FoundNumber: boolean;
        DotPos, I, LineNumber: integer;
        AnnotationLine: string;

      begin (* tfmGCode.lbxAnnotatedClick *)
        with lbxAnnotated do
          begin (* lbxAnnotated *)
            I := ItemIndex;
            LineNumber := 1;
            FoundNumber := false;

            repeat (* seek numbered line *)
              AnnotationLine := Trim(Items[I]);
              if length(AnnotationLine) > 0 then
                case AnnotationLine[1] of
                  '0' .. '9': begin (* found *)
                                DotPos := pos(PERIOD, AnnotationLine);
                                if DotPos > 0 then
                                  try    (* conversion is fragile *)
                                    LineNumber := StrToInt(Copy(AnnotationLine, 1, DotPos-1));
                                    FoundNumber := true;
                                    break
                                  except (* conversion is fragile *)
                                    on E: EConvertError do
                                      LineNumber := 1
                                  end    (* conversion is fragile *)
                              end   (* found *);
                         else if integer(Items.Objects[I]) <> 0 then
                                begin (* Oh-number message *)
                                  LineNumber := integer(Items.Objects[I]);
                                  FoundNumber := true;
                                  break
                                end   (* Oh-number message *)
                  end (* case *);
              dec(I)
            until  (* seek numbered line *) FoundNumber or (I < 0);

            if FoundNumber then
              with memoOriginal do
               begin (* scroll to position *)
                 Perform(EM_LINESCROLL, 0, -9999);
                 Perform(EM_LINESCROLL, 0, LineNumber-1)
               end   (* scroll to position *)
          end   (* lbxAnnotated *)
      end   (* tfmGCode.lbxAnnotatedClick *);

    procedure tfmGCode.MakeAnnotation(const A: string; LineNumber: integer);
      (* A describes bad syntax here *)

      begin (* tfmGCode.MakeAnnotation *)
        ErrorsFound := true;
        with lbxAnnotated do
          begin (* show line if not already shown *)
            if length(Trim(LineOfInterest)) > 6 then
              begin (* show parse elements too *)
                ItemIndex := Items.Add(LineOfInterest);
                if ckbxDebug.Checked then
                  ItemIndex := Items.Add('   ' + Parsed)
              end   (* show parse elements too *);

            LineOfInterest := nullstring;
            ItemIndex := Items.AddObject('   ' + A, tObject(LineNumber))
          end   (* show line if not already shown *)
      end   (* tfmGCode.MakeAnnotation *);

{************ variables *****************************************************}

    procedure tfmGCode.btnSaveVarsToFileClick(Sender: tObject);
      (* save contents of list boxes *)

      var
        SaveExt, SaveName: string;
        TheFile: textfile;

      procedure SaveOneBox(TheBox: tListBox; const S: string);
        (* save the box with the legend *)

        var
          I: integer;

        begin (* SaveOneBox *)
          writeln(TheFile, S);
          with TheBox do
            if Count = 0 then
              writeln(TheFile, '  (none)')
            else
              for I := 1 to Count do
                writeln(TheFile, '  ' + Items[I-1]);
          writeln(TheFile, nullstring);
        end   (* SaveOneBox *);

      begin (* tfmGCode.btnSaveVarsToFileClick *)
        SaveName := edFileName.Text;
        SaveExt := ExtractFileExt(edFileName.Text);
        SaveName := copy(SaveName, 1, length(SaveName) - length(SaveExt));
        SaveName := SaveName + '.vars.txt';
        with tSaveDialog.Create(nil) do
          try     (* must free dialog *)
            DefaultExt := 'txt';
            FileName := SaveName;
            Title := 'save list boxes to file';
            if Execute then
              begin (* write the file *)
                AssignFile(TheFile, FileName);
                rewrite(TheFile);

                SaveOneBox(lbxVarsAssigned,          'variables assigned');
                SaveOneBox(lbxVarsUsed,              'variables referenced');
                SaveOneBox(lbxAssignedNoReference,   'assigned but not referenced');
                SaveOneBox(lbxReferencedNotAssigned, 'referenced nut not assigned');
                CloseFile(TheFile)
              end   (* write the file *)
          finally (* must free dialog *)
            Free
          end     (* must free dialog *);
      end   (* tfmGCode.btnSaveVarsToFileClick *);

    procedure tfmGCode.LogVariable(const VarName: string; const OnLHS: boolean);
      (* add this to the appropriate listbox *)

      var
        Obsolete: integer;

      begin (* tfmGCode.LogVariable *)
        if OnLHS then
          begin (* assigned to *)
            with lbxVarsAssigned.Items do
              if IndexOf(VarName) < 0 then
                Add(VarName);
            if lbxVarsUsed.Items.IndexOf(VarName) < 0 then
              with lbxAssignedNoReference.Items do
                if IndexOf(VarName) < 0 then
                  Add(VarName);
            Obsolete := lbxReferencedNotAssigned.Items.IndexOf(VarName);
            if Obsolete >= 0 then
              lbxReferencedNotAssigned.Items.Delete(Obsolete)
          end   (* assigned to *)

        else
          begin (* referenced *)
            with lbxVarsUsed.Items do
              if IndexOf(VarName) < 0 then
                Add(VarName);
            if lbxVarsAssigned.Items.IndexOf(VarName) < 0 then
              with lbxReferencedNotAssigned.Items do
                if IndexOf(VarName) < 0 then
                  Add(VarName);
            Obsolete := lbxAssignedNoReference.Items.IndexOf(VarName);
            if Obsolete >= 0 then
              lbxAssignedNoReference.Items.Delete(Obsolete)
          end   (* referenced *)
      end   (* tfmGCode.LogVariable *);

end.
